@@ -1,5 +1,5 @@
 /*  GnomeKiss - A KiSS viewer for the GNOME desktop
-    Copyright (C) 2000-2001  Nick Lamb <njl195@zepler.org.uk>
+    Copyright (C) 2000-2002  Nick Lamb <njl195@zepler.org.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,15 +21,16 @@
 #endif
 
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <unistd.h>
 #include <gnome.h>
 #include "support.h"
 #include "kiss.h"
 
-KissActionList errors;
+KissActionList fkiss_errors;
 
 static int random_value(int from, int to);
-static KissActionList label_id(guint id);
+static KissActionList label_by_id(gint id);
 static void object_unmap(KissObject *object);
 static void object_map(KissObject *object);
 static void object_altmap(KissObject *object);
@@ -37,9 +38,12 @@ static void object_ghost(KissObject *object, int value);
 static void object_transparent(KissObject *object, int amount);
 static void object_move_relative(KissObject *object, int x, int y);
 static void play_midi(gchar *song);
+static void stop_midi(void);
 
-static unsigned int depth= 0;
-static unsigned int control_flag[16];
+static unsigned int depth= 0, cycles = 0;
+static unsigned int control_flag[MAX_DEPTH+1];
+
+static pid_t midi_child= 0; /* MIDI child process */
 
 /* returns a random integer between "from" and "to" inclusive */
 static int random_value(int from, int to) {
@@ -58,12 +62,16 @@ static int random_value(int from, int to) {
   return from + (rand() % (range + 1));
 }
 
-static KissActionList label_id(guint id) {
-  if (id >= LABELS || config.labels[id] == NULL) {
-    log_error(_("label(%u) is undefined"), id);
+static KissActionList label_by_id(gint id) {
+  KissActionList *event;
+
+  event = event_find(config.labels, id);
+
+  if (event == NULL) {
+    log_error(_("label(%d) is undefined"), id);
     return NULL;
   } else {
-    return config.labels[id];
+    return *event;
   }
 }
 
@@ -74,7 +82,6 @@ static void object_unmap(KissObject *object) {
     ((KissCell *) list->data)->mapped= 0;
   }
   render_object(object);
-  check_collisions(object, 1);
 }
 
 static void object_map(KissObject *object) {
@@ -84,7 +91,6 @@ static void object_map(KissObject *object) {
     ((KissCell *) list->data)->mapped= 1;
   }
   render_object(object);
-  check_collisions(object, 1);
 }
 
 static void object_altmap(KissObject *object) {
@@ -94,7 +100,6 @@ static void object_altmap(KissObject *object) {
     ((KissCell *) list->data)->mapped^= 1;
   }
   render_object(object);
-  check_collisions(object, 1);
 }
 
 static void object_ghost(KissObject *object, int value) {
@@ -122,18 +127,45 @@ static void object_move_relative(KissObject *object, int x, int y) {
   object_move(object, object->x[view] + x, object->y[view] + y);
 }
 
-static void play_midi(gchar *song) {
-  pid_t child;
+static void stop_midi(void) {
 
   if (!prefs.midi) return; /* user doesn't want MIDI */
 
-  child = fork();
-  if (child == 0) {
-    execlp(prefs.midi_player, prefs.midi_player, song, NULL);
-    _exit(182);
+  if (midi_child) {
+    kill(midi_child, SIGTERM);
+    waitpid(midi_child, NULL, 0);
+    midi_child= 0;
   }
 }
 
+static void play_midi(gchar *song) {
+  int k;
+  char *string, *args[13]; /* 1 command, 10 parameters, 1 song, 1 NULL */
+
+  if (!prefs.midi) return; /* user doesn't want MIDI */
+
+  string = strdup(prefs.midi_player);
+  args[0]= strtok(string, " ");
+  for (k= 1; k < 11; ++k) {
+    args[k]= strtok(NULL, " ");
+    if (args[k] == NULL) break;
+  }
+  args[k]= song;
+  args[k+1]= NULL;
+
+  stop_midi(); /* just in case */
+
+  midi_child = fork();
+  if (midi_child == 0) {
+    execvp(args[0], args);
+    _exit(182);
+  } else if (midi_child == -1) {
+    midi_child = 0;
+  }
+
+  free(string);
+}
+
 static void trace(KissAction *action, int depth) {
   int k;
 
@@ -141,12 +173,13 @@ static void trace(KissAction *action, int depth) {
 
   if (!prefs.trace) return;
 
+  fprintf(stderr, "%4u ", action->line);
   for(k= 1; k < depth; ++k) {
     fputc(' ', stderr);
   }
 
   if (action->type < CONTROL_CONDITIONAL && control_flag[depth] == 0) {
-    fprintf(stderr, _("(skipped) "));
+    fprintf(stderr, _("##skipped## "));
   }
 
   switch (action->type) {
@@ -578,7 +611,7 @@ static void trace(KissAction *action, int depth) {
     break;
 
   default:
-    fprintf(stderr, "(UNKNOWN)\n");
+    fprintf(stderr, "## Unknown action? ##\n");
   }
 }
 
@@ -604,12 +637,10 @@ static KissActionList trigger(KissAction *action, KissActionList next) {
     render_object(KissObject(action->args[0]));
     break;
   case ACTION_CHANGESET:
-    switch_view(KissValue(action->args[0]));
-    render_all();
+    switch_view(KissValue(action->args[0])); /* includes render */
     break;
   case ACTION_CHANGECOL:
-    switch_color(KissValue(action->args[0]));
-    render_all();
+    switch_color(KissValue(action->args[0])); /* includes render */
     break;
   case ACTION_DEBUG:
     fprintf(stderr, _("DEBUG: %s\n"), KissString(action->args[0]));
@@ -648,7 +679,7 @@ static KissActionList trigger(KissAction *action, KissActionList next) {
     break;
   case ACTION_SOUND_CANCEL:
     if (prefs.esound) {
-      /* FIXME: use esd_sample_kill(gnome_sound_connection, id) */
+      /* FIXME: it's not clear if even a non-portable cancel exists (!) */
       fprintf(stderr, "Don't know how to cancel sounds yet\n");
     }
     break;
@@ -703,9 +734,7 @@ static KissActionList trigger(KissAction *action, KissActionList next) {
     play_midi(KissString(action->args[0]));
     break;
   case ACTION_MUSIC_CANCEL:
-    if (prefs.midi) {
-      fprintf(stderr, _("Sorry cannot cancel music yet\n"));
-    }
+    stop_midi();
     break;
   case ACTION_NOTIFY:
     gnome_app_message(GNOME_APP(app), KissString(action->args[0]));
@@ -852,18 +881,21 @@ static KissActionList trigger(KissAction *action, KissActionList next) {
     break;
 
   case CONTROL_GOTO:
-    next= label_id(KissValue(action->args[0]));
+    next= label_by_id(KissValue(action->args[0]));
     break;
     
   case CONTROL_GOSUB:
-    events(label_id(KissValue(action->args[0])));
+    if (next == NULL) /* tail recursion == loop, avoid stack */
+      next= label_by_id(KissValue(action->args[0]));
+    else
+      events(label_by_id(KissValue(action->args[0])));
     break;
     
   case CONTROL_GOTO_RANDOM:
     if (random_value(1, 100) <= KissValue(action->args[0])) {
-      next= label_id(KissValue(action->args[1]));
+      next= label_by_id(KissValue(action->args[1]));
     } else {
-      next= label_id(KissValue(action->args[2]));
+      next= label_by_id(KissValue(action->args[2]));
     }
     break;
 
@@ -873,9 +905,9 @@ static KissActionList trigger(KissAction *action, KissActionList next) {
     else
       k = 2;
     if (next == NULL) /* tail recursion == loop, avoid stack */
-      next= label_id(KissValue(action->args[k]));
+      next= label_by_id(KissValue(action->args[k]));
     else
-      events(label_id(KissValue(action->args[k])));
+      events(label_by_id(KissValue(action->args[k])));
     break;
     
   case CONTROL_EQUAL:
@@ -919,7 +951,7 @@ static KissActionList trigger(KissAction *action, KissActionList next) {
 void events(KissActionList list) {
   if (list == NULL) return;
 
-  if (depth > 5) {
+  if (depth >= MAX_DEPTH) {
     fprintf(stderr, _("French KiSS stack overflow.\n"));
     return;
   } else {
@@ -929,7 +961,12 @@ void events(KissActionList list) {
   while (list != NULL) {
     trace(list->data, depth);
     list= trigger(list->data, g_slist_next(list));
+    if (++cycles % 4096 == 0 && gtk_main_iteration_do(FALSE))
+      exit(0); /* all done */
   }
   depth--;
+  if (depth == 0) {
+    cycles = 0; /* reset */
+  }
 }
 
