@@ -1,5 +1,5 @@
 /*  GnomeKiss - A KiSS viewer for the GNOME desktop
-    Copyright (C) 2000-2001  Nick Lamb <njl195@zepler.org.uk>
+    Copyright (C) 2000-2002  Nick Lamb <njl195@zepler.org.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -27,9 +27,7 @@
 
 #include "kiss.h"
 
-#define NUMERIC "%[A-Za-z0-9-]"
-#define OBJID "#%4u"
-#define STRING "\"%[^\"]\""
+#define FK_ERRORS &fkiss_errors
 
 static int parse(char *args, char *argv[], char types[], int max);
 static KissActionList *find_event(gchar *event, gchar *params);
@@ -49,7 +47,7 @@ void reset_french() {
   }
   action_chunk= g_mem_chunk_create(KissAction, 100, G_ALLOC_ONLY);
 
-  script= ERRORS;
+  script= NULL;
 }
 
 static int parse(char *args, char *argv[], char types[], int max) {
@@ -129,27 +127,27 @@ static KissActionList *find_event(gchar *event, gchar *params) {
   if (!strcmp("alarm", event)) {
     if (count == 1 && types[0] == '?') {
       val= numeric(first);
-      if (val >= 0 && val < ALARMS) {
-        return &(config.alarms[val]);
+      if (val >= 0 && val <= INT_MAX) {
+        return event_find_or_insert(config.alarms, val);
       } else {
         parse_error(event, params);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("apart", event)) { /* FKiSS 2.1 */
     if (count == 2 && types[0] == 'S' && types[1] == 'S') {
       cell_hit = collision_find(first, second);
       if (cell_hit == NULL)
-        return ERRORS; /* Reason logged in collision_find()  */
+        return FK_ERRORS; /* Reason logged in collision_find()  */
       else
         return &(cell_hit->apart);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("begin", event)) {
@@ -157,7 +155,7 @@ static KissActionList *find_event(gchar *event, gchar *params) {
       return &(config.begin);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("catch", event)) {
@@ -166,18 +164,18 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->catch;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->catch);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("col", event)) {
@@ -187,23 +185,23 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &(config.cols[val]);
       } else {
         parse_error(event, params);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("collide", event)) { /* FKiSS 2.1 */
     if (count == 2 && types[0] == 'S' && types[1] == 'S') {
       cell_hit = collision_find(first, second);
       if (cell_hit == NULL)
-        return ERRORS; /* Reason logged in collision_find()  */
+        return FK_ERRORS; /* Reason logged in collision_find()  */
       else
         return &(cell_hit->collide);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("drop", event)) {
@@ -212,18 +210,18 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->drop;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->drop);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("end", event)) {
@@ -231,7 +229,7 @@ static KissActionList *find_event(gchar *event, gchar *params) {
       return &(config.end);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("fixcatch", event)) {
@@ -240,18 +238,18 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->fixcatch;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->fixcatch);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("fixdrop", event)) {
@@ -260,30 +258,30 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->fixdrop;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->fixdrop);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("in", event)) { /* FKiSS 2 */
     if (count == 2 && types[0] == 'O' && types[1] == 'O') {
       object_hit = impact_find(first, second);
       if (object_hit == NULL)
-        return ERRORS; /* Reason logged in impact_find()  */
+        return FK_ERRORS; /* Reason logged in impact_find()  */
       else
         return &(object_hit->in);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("initialize", event)) {
@@ -291,21 +289,21 @@ static KissActionList *find_event(gchar *event, gchar *params) {
       return &(config.init);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("label", event)) { /* FKiSS 3 */
     if (count == 1 && types[0] == '?') {
       val= numeric(first);
-      if (val >= 0 && val < LABELS) {
-        return &(config.labels[val]);
+      if (val >= 0 && val <= INT_MAX) {
+        return event_find_or_insert(config.labels, val);
       } else {
         parse_error(event, params);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("never", event)) {
@@ -313,19 +311,19 @@ static KissActionList *find_event(gchar *event, gchar *params) {
       return &(config.never);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("out", event)) { /* FKiSS 2 */
     if (count == 2 && types[0] == 'O' && types[1] == 'O') {
       object_hit = impact_find(first, second);
       if (object_hit == NULL)
-        return ERRORS; /* Reason logged in impact_find()  */
+        return FK_ERRORS; /* Reason logged in impact_find()  */
       else
         return &(object_hit->out);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("overflow", event)) /* FKiSS 3 */ {
@@ -333,7 +331,7 @@ static KissActionList *find_event(gchar *event, gchar *params) {
       return &(config.overflow);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("press", event)) {
@@ -342,18 +340,18 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->press;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->press);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("release", event)) {
@@ -362,18 +360,18 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->release;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->release);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
 
@@ -384,35 +382,35 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &(config.sets[val]);
       } else {
         parse_error(event, params);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("stillin", event)) { /* FKiSS 2 */
     if (count == 2 && types[0] == 'O' && types[1] == 'O') {
       object_hit = impact_find(first, second);
       if (object_hit == NULL)
-        return ERRORS; /* Reason logged in impact_find()  */
+        return FK_ERRORS; /* Reason logged in impact_find()  */
       else
         return &(object_hit->stillin);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("stillout", event)) { /* FKiSS 2 */
     if (count == 2 && types[0] == 'O' && types[1] == 'O') {
       object_hit = impact_find(first, second);
       if (object_hit == NULL)
-        return ERRORS; /* Reason logged in impact_find()  */
+        return FK_ERRORS; /* Reason logged in impact_find()  */
       else
         return &(object_hit->stillout);
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("unfix", event)) {
@@ -421,18 +419,18 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &cell->events->unfix;
       } else {
         log_error(_("cell \"%s\" not loaded"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else if (count == 1 && types[0] == 'O') {
       if ((object= object_find(first)) != NULL) {
         return &(object->unfix);
       } else {
         log_error(_("invalid object #%s"), first);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   } else if (!strcmp("version", event)) { /* FKiSS 2 */
@@ -441,11 +439,11 @@ static KissActionList *find_event(gchar *event, gchar *params) {
         return &(config.version);       /* We claim FKiSS 3 compliance */
       } else {
         log_warning(_("FKiSS version %s not available"), params);
-        return ERRORS;
+        return FK_ERRORS;
       }
     } else {
       parse_error(event, params);
-      return ERRORS;
+      return FK_ERRORS;
     }
 
   }
@@ -459,6 +457,7 @@ static KissAction *new_action(KissActionType type) {
 
   action= g_chunk_new0(KissAction, action_chunk);
   action->type= type;
+  action->line= get_line();
 
   return action;
 }
@@ -487,7 +486,9 @@ static void *int_helper(gchar *integer, long min, long max) {
 static void check_and_add(char *method, char *params, KissAction *action,
                                                              int required) {
 
-  if (required > 0 && action->args[0] == NULL) {
+  if (script == NULL) {
+    log_error(_("%s(%s) actions must be inside an event"), method, params);
+  } else if (required > 0 && action->args[0] == NULL) {
     log_error(_("%s(%s) first parameter incorrect"), method, params);
   } else if (required > 1 && action->args[1] == NULL) {
     log_error(_("%s(%s) second parameter incorrect"), method, params);
@@ -637,7 +638,7 @@ void parse_action(guchar *text) {
   } else if (!strcmp("gosub", method)) {
     if (count == 1 && types[0] == '?') {
       action= new_action(CONTROL_GOSUB);
-      action->args[0]= int_helper(first, 0, LABELS - 1);
+      action->args[0]= int_helper(first, 0, INT_MAX);
       check_and_add(method, params, action, 1);
     } else {
       parse_error(method, params);
@@ -647,8 +648,8 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == '?' && types[1] == '?' && types[2] == '?') {
       action= new_action(CONTROL_GOSUB_RANDOM);
       action->args[0]= int_helper(first, 0, 100);
-      action->args[1]= int_helper(second, 0, LABELS - 1);
-      action->args[2]= int_helper(third, 0, LABELS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
+      action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
       parse_error(method, params);
@@ -657,7 +658,7 @@ void parse_action(guchar *text) {
   } else if (!strcmp("goto", method)) {
     if (count == 1 && types[0] == '?') {
       action= new_action(CONTROL_GOTO);
-      action->args[0]= int_helper(first, 0, LABELS - 1);
+      action->args[0]= int_helper(first, 0, INT_MAX);
       check_and_add(method, params, action, 1);
     } else {
       parse_error(method, params);
@@ -667,8 +668,8 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == '?' && types[1] == '?' && types[2] == '?') {
       action= new_action(CONTROL_GOTO_RANDOM);
       action->args[0]= int_helper(first, 0, 100);
-      action->args[1]= int_helper(second, 0, LABELS - 1);
-      action->args[2]= int_helper(third, 0, LABELS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
+      action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
       parse_error(method, params);
@@ -690,7 +691,7 @@ void parse_action(guchar *text) {
 
       action= new_action(ACTION_IF_FIXED);
       action->args[0]= (void *) object_find(first);
-      action->args[1]= int_helper(second, 0, ALARMS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
@@ -721,7 +722,7 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == 'S' && types[1] == '?' && types[2] == '?') {
       action= new_action(ACTION_IF_MAPPED);
       action->args[0]= (void *) cell_find(first);
-      action->args[1]= int_helper(second, 0, ALARMS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
@@ -732,7 +733,7 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == 'O' && types[1] == '?' && types[2] == '?') {
       action= new_action(ACTION_IF_MOVED);
       action->args[0]= (void *) object_find(first);
-      action->args[1]= int_helper(second, 0, ALARMS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
@@ -753,7 +754,7 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == 'O' && types[1] == '?' && types[2] == '?') {
       action= new_action(ACTION_IF_NOT_FIXED);
       action->args[0]= (void *) object_find(first);
-      action->args[1]= int_helper(second, 0, ALARMS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
@@ -764,7 +765,7 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == 'S' && types[1] == '?' && types[2] == '?') {
       action= new_action(ACTION_IF_NOT_MAPPED);
       action->args[0]= (void *) cell_find(first);
-      action->args[1]= int_helper(second, 0, ALARMS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
@@ -775,7 +776,7 @@ void parse_action(guchar *text) {
     if (count == 3 && types[0] == 'O' && types[1] == '?' && types[2] == '?') {
       action= new_action(ACTION_IF_NOT_MOVED);
       action->args[0]= (void *) object_find(first);
-      action->args[1]= int_helper(second, 0, ALARMS - 1);
+      action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
     } else {
@@ -1073,7 +1074,7 @@ void parse_action(guchar *text) {
   } else if (!strcmp("randomtimer", method)) {
     if (count == 3 && types[0] == '?' && types[1] == '?' && types[2] == '?') {
       action= new_action(ACTION_RANDOMTIMER);
-      action->args[0]= int_helper(first, 0, ALARMS - 1);
+      action->args[0]= int_helper(first, 0, INT_MAX);
       action->args[1]= int_helper(second, 0, INT_MAX);
       action->args[2]= int_helper(third, 0, INT_MAX);
       check_and_add(method, params, action, 3);
@@ -1129,7 +1130,7 @@ void parse_action(guchar *text) {
   } else if (!strcmp("timer", method)) {
     if (count == 2 && types[0] == '?' && types[1] == '?') {
       action= new_action(ACTION_TIMER);
-      action->args[0]= int_helper(first, 0, ALARMS - 1);
+      action->args[0]= int_helper(first, 0, INT_MAX);
       action->args[1]= int_helper(second, 0, INT_MAX);
       check_and_add(method, params, action, 2);
     } else {
@@ -1189,7 +1190,7 @@ void parse_action(guchar *text) {
 
     if (event != NULL) {
       script= event;
-      if (event != ERRORS && *event != NULL) {
+      if (event != FK_ERRORS && *event != NULL) {
         log_warning(_("%s(%s) event used more than once"), method, params);
       }
     }
